#include "gba.h"
#include <stdio.h>
#include <varargs.h>

#define SIO_ACK 0x0008;

void rx_wait(u32 w,void *buf){
	u32 i;
	for(i=0;i<w;i++){
		*(vu16 *)REG_SIOCNT = SIO_32BIT_MODE;
		*(vu16 *)REG_SIOCNT = SIO_32BIT_MODE | SIO_START | SIO_ACK;
		for(;;){
			if((*(vu16 *)REG_SIOCNT & SIO_START)==0) break;
		}
		*((vu16 *)buf+i*2+0)=*((vu16 *)REG_SIODATA32+0);
		*((vu16 *)buf+i*2+1)=*((vu16 *)REG_SIODATA32+1);
	}
}

void tx_wait(u32 w,void *buf){
	u32 i;
	for(i=0;i<w;i++){
		*((vu16 *)REG_SIODATA32+0)=*((vu16 *)buf+i*2+0);
		*((vu16 *)REG_SIODATA32+1)=*((vu16 *)buf+i*2+1);
		*(vu16 *)REG_SIOCNT = SIO_32BIT_MODE;
		*(vu16 *)REG_SIOCNT = SIO_32BIT_MODE | SIO_START | SIO_ACK;
		for(;;){
			if((*(vu16 *)REG_SIOCNT & SIO_START)==0) break;
		}
	}
}

s32 PcOpen(s8 *filename,s32 mode){
	u32 i;
	u32 buf[64];
	u32 len;
	buf[0]=0xf0f0f001;
	for(len=0;;len++) if(filename[len]==0) break;
	len=(strlen(filename)+1+3)/4;
	buf[1]=len;
	for(i=0;i<len;i++){
		buf[2+i]=*((u32 *)(filename+i*4));
	}
	buf[2+len]=mode;
	tx_wait(3+len,buf);
	rx_wait(1,&len);
	return(len);
}

s32 PcClose(s32 fh){
	u32 buf[2];
	buf[0]=0xf0f0f002;
	buf[1]=(u32)fh;
	tx_wait(2,buf);
	return(0);
}

s32 PcRead(s32 fh,void *buffer,s32 size){
	s32 len;
	s32 zan,ofs;
	u32 buf[64];
	buf[0]=0xf0f0f003;
	buf[1]=(u32)fh;
	buf[2]=(u32)size;
	tx_wait(3,buf);
	rx_wait(1,&size);
	if(size>0){
		zan=size;
		ofs=0;
		for(;;){
			if(zan>256){
				len=256;
			}
			else{
				len=zan;
			}
			rx_wait((len+3)/4,buf);
			DmaCopy(3,buf,(u32)buffer+ofs,(len+3)/4*4,16);
			ofs+=len;
			zan-=len;
			if(zan==0) break;
		}
	}
	return(size);
}

s32 PcWrite(s32 fh,void *buffer,s32 size){
	u32 buf[64];
	s32 len,zan,ofs;
	zan=size;
	ofs=0;
	for(;;){
		if(zan>256){
			len=256;
		}
		else{
			len=zan;
		}
		buf[0]=0xf0f0f004;
		buf[1]=(u32)fh;
		buf[2]=(u32)len;
		tx_wait(3,buf);
		DmaCopy(3,(u32)buffer+ofs,buf,(len+3)/4*4,16);
		tx_wait((len+3)/4,buf);
		ofs+=len;
		zan-=len;
		if(zan==0) break;
	}
	return(size);
}

s32 PcLseek(s32 fh,s32 offset,s32 where){
	u32 buf[4];
	buf[0]=0xf0f0f005;
	buf[1]=(u32)fh;
	buf[2]=(u32)offset;
	buf[3]=(u32)where;
	tx_wait(4,buf);
	rx_wait(1,buf);
	return((s32)buf[0]);
}

static s32 _udeci(u32 u,s8 *s,s32 i){
	u32 div,flg;
	div=1000000000;
	flg=0;
	for(;;){
		if(u/div){
			s[i++]='0'+(u/div);
			flg=1;
			u-=(u/div)*div;
		}
		else if(flg || (div==1)){
			s[i++]='0';
		}
		div=div/10;
		if(div==0) break;
	}
	return(i);
}

const static s8 hexchar[16]={
'0','1','2','3','4','5','6','7',
'8','9','A','B','C','D','E','F'
};

static s32 _hex(u32 u,s8 *s,s32 i){
	u32 div,flg;
	div=0x10000000;
	flg=0;
	for(;;){
		if(u/div){
			s[i++]=hexchar[u/div];
			flg=1;
			u-=(u/div)*div;
		}
		else if(flg || (div==1)){
			s[i++]='0';
		}
		div=div/16;
		if(div==0) break;
	}
	return(i);
}

//	%d:POi
//	%u:ȂPOi
//	%x:PUi
//	%s:
//	%c:
//	\n:s

s32 PcPrintf(va_alist) va_dcl{
	va_list args;
	s32 i,d;
	u32 u;
	s8 *ss;
	s8 *fmt;
	s8 *s;
	u32 buf[128];

	buf[0]=0xf0f0f000;
	s=(u8 *)(buf+2);
	va_start(args);
	fmt=va_arg(args,char *);
	for(i=0;;){
		if(*fmt=='%'){
			switch(fmt[1]){
			case 'd':		//	%d:POi
				fmt+=2;
				d=va_arg(args,s32);
				if(d<0){
					s[i++]='-';
					u=-d;
				}
				else{
					u=d;
				}
				i=_udeci(u,s,i);
				break;
			case 'u':		//	%u:ȂPOi
				fmt+=2;
				u=va_arg(args,u32);
				i=_udeci(u,s,i);
				break;
			case 'x':		//	%x:PUi
				fmt+=2;
				u=va_arg(args,u32);
				i=_hex(u,s,i);
				break;
			case 's':		//	%s:
				fmt+=2;
				ss=va_arg(args,s8 *);
				for(d=0;;d++){
					if(ss[d]==0) break;
					s[i++]=ss[d];
				}
				break;
			case 'c':		//	%c:
				fmt+=2;
				d=va_arg(args,s32);
				s[i++]=d;
				break;
			default:
				s[i++]=*fmt;
				fmt++;
				break;
			}
		}
		else{
			if((i>240)||(*fmt==0)){
				s[i++]=0;
				break;
			}
			s[i++]=*fmt;
			fmt++;
		}
	}
	va_end(args);
	buf[1]=(i+3)/4;
	tx_wait(2+(i+3)/4,buf);
	return(0);
}
